import configparser
import re


from karyoplot.entities import Sequence, Canvas, Track


# Class Confiparser
class Confparser(configparser.ConfigParser):
    def __init__(self):
        configparser.ConfigParser.__init__(self)

    # Lists all the sections starting with "track_"
    def find_track(self):
        tracks = []
        for section in self.sections():
            if re.search("^track_(\S)+$", section):
                tracks.append(section)
        return tracks

    # Returns a list of track Objects
    def get_tracks(self):
        track_list = []
        for i in self.find_track():
            kwargs = dict(self.items(i))
            t = Track(**kwargs)
            track_list.append(t)
        return track_list

    # Returns a canvas Object
    def get_canvas(self):
        kwargs = dict(self.items("canvas"))
        c = Canvas(**kwargs)
        return c

    def get_sequence(self):
        kwargs = dict(self.items("sequence"))
        s = Sequence(**kwargs)
        return s

    # Returns zoom on sequences
    # list of tuples (chr,start,end)
    # start and end could be None
    def get_zooms(self):
        kwargs = dict(self.items("sequence"))
        if 'zoom' in kwargs:
            seqs = kwargs['zoom'].split(",")
            for seq in seqs:
                m = re.match(r"(\w+)(:)*(\d+)*(-)*(\d+)*", seq)
                if m:
                    if m.group(3) and m.group(5):
                        yield (m.group(1), int(m.group(3)), int(m.group(5)))
                    else:
                        yield (m.group(1), None, None)
        else:
            return []




import pysam

from matplotlib import collections as mc
import matplotlib.pyplot as plt

class Sequence():
    def __init__(self, *args, **kwargs):

        self.kwargs = kwargs

        self.zoom = Utils.check_kwargs(self.kwargs,'zoom', str, '')
        self.size = Utils.check_kwargs(self.kwargs,'size', int, 1)
        self.color = Utils.check_kwargs(self.kwargs,'color', str, 'black')
        self.limitesize = Utils.check_kwargs(self.kwargs,'limitesize', int, 50)
        self.reference = Utils.check_kwargs(self.kwargs,'reference', str, '')
        self.scale = Utils.check_kwargs(self.kwargs,'scale', bool, False)
        self.seq_fontsize = Utils.check_kwargs(self.kwargs,'seq_fontsize', int, 14)
        self.x_label_fontsize = Utils.check_kwargs(self.kwargs, 'x_label_fontsize', int, 20)
        self.sharex = True
        if self.zoom:
            self.sharex = False


    def validate(self):
        if pysam.FastaFile(self.reference):
            return True


class Chromosome:

    def __init__(self, name, length):
        self.name = name
        self.length = length
        self.zoom_min = 0
        self.zoom_max = length


class Track:
    def __init__(self, *args, **kwargs):

        self.kwargs = kwargs

        self.title = Utils.check_kwargs(self.kwargs,'title', str, '')
        self.graph_type = Utils.check_kwargs(self.kwargs,'graph_type', str, '')
        self.size = Utils.check_kwargs(self.kwargs,'size', float, 1)
        self.thickness = Utils.check_kwargs(self.kwargs,'thickness', float, 1)
        self.color = Utils.check_kwargs(self.kwargs,'color', str, 'red')
        self.data = Utils.check_kwargs(self.kwargs,'data', str, '')
        self.datatype = Utils.check_kwargs(self.kwargs,'datatype', str, '')
        self.background_color = Utils.check_kwargs(self.kwargs,'background_color', str, 'white')
        self.GC_window_length = Utils.check_kwargs(self.kwargs,'gc_window_length', int, 1000)
        self.GC_window_overlap = Utils.check_kwargs(self.kwargs,'gc_window_overlap', int, 100)
        self.scaley = Utils.check_kwargs(self.kwargs,'scaley', bool, False)
        self.y_label_fontsize = Utils.check_kwargs(self.kwargs, 'y_label_fontsize', int, 20)



    def draw_track(self, sequence, ax, chrom, sharex=True):

        if not sharex:
            ax.set_xlim([chrom.zoom_min, chrom.zoom_max])
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.get_xaxis().set_visible(False)

        # Parsed depending on datatype option
        if self.datatype.lower() == "bed" or self.datatype.lower() == "gff":
            tbx = pysam.TabixFile(self.data)
            coordinates = []
            # Bed et GFF both parsed as Tuples
            color = "black"
            for file in tbx.fetch(chrom.name, chrom.zoom_min, chrom.zoom_max, parser=pysam.asTuple()):
                coordinates.append([[file[1], 1], [file[2], 1]])
                color = file[4]

            lc = mc.LineCollection(coordinates, colors=str(color), linewidths=self.thickness)
            ax.add_collection(lc)
            ax.set_ylim([0, 2])


        if self.datatype.lower() == "vcf":
            tbx = pysam.TabixFile(self.data)
            coordinates = []
            for vcf in tbx.fetch(chrom.name, chrom.zoom_min, chrom.zoom_max, parser=pysam.asVCF()):
                coordinates.append([vcf[1], 1])

            lc = mc.LineCollection(coordinates, colors=self.color, linewidths=self.thickness)
            ax.add_collection(lc)
            ax.set_ylim([0, 2])

        # If type is gc a sliding window must be called
        gc_min = []
        gc_max = []
        if self.graph_type.lower() == "gc":
            base_start, gcContent = self.__GC(sequence, chrom)
            ax.plot(base_start, gcContent, c=self.color, linewidth=self.thickness)

            gc_max.append(max(gcContent))
            gc_min.append(min(gcContent))

            if self.scaley:
                ax.set_ylim([min(gc_min) - 2, max(gc_max) + 2])
                ax.spines['left'].set_visible(True)
                ax.get_yaxis().set_visible(True)
                ax.tick_params(axis='y', labelsize=self.y_label_fontsize)

        # re-add background limited to max x
        # axis off, turned off set_facecolor
        # see: https://stackoverflow.com/questions/60805253/matplotlib-turning-axes-off-and-setting-facecolor-at-the-same-time-not-possible
        #ax.add_patch(plt.Rectangle((0, -5), chrom.length, 100, facecolor=self.background_color, zorder=-100))
        ax.add_patch(plt.Rectangle((chrom.zoom_min, -5), chrom.zoom_max-chrom.zoom_min, 100, facecolor=self.background_color, zorder=-100))

    def __GC(self, sequence, chrom):

        fasta_file = pysam.FastaFile(sequence.reference)
        base_start = []
        gcContent = []
        step = self.GC_window_length - self.GC_window_overlap
        seq = fasta_file.fetch(chrom.name)[chrom.zoom_min:chrom.zoom_max].lower()
        for i in range(chrom.zoom_min, chrom.zoom_max, step):
            idx = i - chrom.zoom_min
            base_start.append(int(i + (step / 2)))  # use middle of window as x
            gc = seq[idx:idx + step].count("g") + seq[idx:idx + step].count("c")
            gcContent.append(gc / step * 100)
        return base_start, gcContent


class Canvas:
    def __init__(self, *args, **kwargs):

        self.kwargs = kwargs

        self.title = Utils.check_kwargs(self.kwargs,'title', str, '')
        self.height = Utils.check_kwargs(self.kwargs,'height', int, 10)
        self.width = Utils.check_kwargs(self.kwargs,'width', int, 10)
        self.background_color = Utils.check_kwargs(self.kwargs,'background_color', str, 'white')
        self.legend = Utils.check_kwargs(self.kwargs,'legend', bool, False)
        self.constrained_layout = Utils.check_kwargs(self.kwargs,'constrained_layout', bool, False)
        self.title_fontsize = Utils.check_kwargs(self.kwargs, 'title_fontsize', int, 20)


class Utils:

    @staticmethod
    def check_kwargs(kwargs, val, t, default):

        if val in kwargs and kwargs[val]:
            if t == int:
                return (int(kwargs[val]))
            if t == str:
                return (str(kwargs[val]))
            if t == float:
                return (float(kwargs[val]))
            if t == bool:
                if kwargs[val].lower() == "true":
                    return True
                if kwargs[val].lower() == "false":
                    return False
        else:
            return default



#!/usr/bin/env python3

import re
import sys
import argparse
import logging
import os.path

import matplotlib.pyplot as plt
from pysam import FastaFile

from karyoplot.confparser import Confparser
from karyoplot.entities import Chromosome


def __add_arguments(parser):
    parser.add_argument(
        "config",
        help="Config File",
        type=str
    )

    parser.add_argument(
        "-v", "--verbosity",
        type=int, choices=[1, 2, 3],
        help="increase output verbosity 1=error, 2=info, 3=debug"
    )

    parser.add_argument(
        "-o", "--output",
        help="Output Image",
        default="karyoplot.png"
    )


def main():
    """script entry point"""
    parser = argparse.ArgumentParser()
    __add_arguments(parser)
    args = parser.parse_args()

    log_level = 'ERROR'
    if args.verbosity == 1:
        log_level = 'ERROR'
    if args.verbosity == 2:
        log_level = 'INFO'
    if args.verbosity == 3:
        log_level = 'DEBUG'
    logging.getLogger().setLevel(log_level)

    # Parse arguments

    # Checking if file exists
    if not os.path.exists(args.config):
        logging.error('CONFIG FILE DOES NOT EXISTS')
        exit(0)
    else:
        logging.info('Creating canvas and track objects')
    cp = Confparser()
    cp.read(args.config)
    canvas = cp.get_canvas()
    tracks = cp.get_tracks()
    sequence = cp.get_sequence()

    ##Plot

    # Reading fasta file
    logging.info('Checking if fasta is valid')
    chromosomes = []
    chrom_idx_name = {}
    if sequence.validate():
        fasta_file = FastaFile(sequence.reference)
        # chr_lengths = {}
        logging.info('Getting sequence lengths')
        for chrom_name in fasta_file.references:
            chrom = Chromosome(chrom_name,
                               fasta_file.get_reference_length(chrom_name))
            chromosomes.append(chrom)
            try:
                chrom_idx_name[chrom.name] = chrom
            except:
                raise Exception("Probably duplicate sequence {}, check your reference file".format(chrom.name))

    # f = open(canvas.reference,'r')
    # print(f.readline())

    nb_chr_to_draw = []
    if not sequence.zoom:
        # Verify if the limitesize is bigger than atleast one sequence
        logging.info('Checking if limitesize is valid')
        # max_length = max(chr_lengths.values())
        max_length = max([chrom.length for chrom in chromosomes])
        if max_length <= sequence.limitesize:
            raise ValueError("The limitesize is too BIG, no Chromosome to draw")

        # List of plots that need to be drawn
        for chrom in chromosomes:
            if chrom.length >= sequence.limitesize:
                nb_chr_to_draw.append(chrom)
    else:
        for zoom in cp.get_zooms():
            if zoom[0] not in [chrom.name for chrom in chromosomes]:
                raise Exception("Error in zoom option, {} is not available in reference file".format(zoom[0]))
            else:
                chrom = Chromosome(zoom[0], fasta_file.get_reference_length(zoom[0]))
                # chrom = chrom_idx_name[zoom[0]]
                if zoom[1]:
                    chrom.zoom_min = min(max(chrom.zoom_min, zoom[1]), chrom.zoom_max)

                if zoom[2]:
                    chrom.zoom_max = max(min(chrom.zoom_max, zoom[2]), chrom.zoom_min)
                nb_chr_to_draw.append(chrom)

    # Blank Canvas
    plt.rcParams["figure.figsize"] = (canvas.width, canvas.height)

    # compute size ratios for each plot (using size option)
    size_ratios = []
    for chrom in nb_chr_to_draw:
        size_ratios.append(sequence.size)
        for track in tracks:
            size_ratios.append(track.size)

    # constrained_layout=True, required to avoid overlap of x-axis
    logging.info("Initializing subplots - quite long")
    fig, ax = plt.subplots(len(nb_chr_to_draw) * (len(tracks) + 1), 1, sharex=sequence.sharex,
                           gridspec_kw={'height_ratios': size_ratios}, constrained_layout=canvas.constrained_layout)

    logging.info('Building Sequences')
    for n, chrom in enumerate(nb_chr_to_draw):
        idx = n * (len(tracks) + 1)
        ax[idx].plot([chrom.zoom_min, chrom.zoom_max], [1, 1], linewidth=5, color=sequence.color)
        ax[idx].get_xaxis().set_visible(True)
        ax[idx].margins(0, 0)

        if not sequence.sharex:
            ax[idx].spines['top'].set_visible(False)
            ax[idx].spines['right'].set_visible(False)
            ax[idx].spines['bottom'].set_visible(False)
            ax[idx].ticklabel_format(axis="x", style="sci", scilimits=(0, 0))
            ax[idx].get_xaxis().set_visible(False)
            if sequence.scale:
                ax[idx].spines['bottom'].set_visible(True)
                ax[idx].get_xaxis().set_visible(True)
            ax[idx].spines['left'].set_visible(False)
            ax[idx].get_yaxis().set_visible(False)
            ax[idx].tick_params(axis='x', labelsize=sequence.x_label_fontsize)



        else:
            ax[idx].axis('off')

        ax[idx].text(0, 0, chrom.name, fontsize=sequence.seq_fontsize, transform=ax[idx].transAxes, ha="right",
                     va="baseline")

        #legend_line = []
        #legend_title = []
        for ix, track in enumerate(tracks):
            #legend_title.append(track.title)
            #legend_line.append(Line2D([0, 1], [0, 1], color=track.color))

            idx_track = idx + (ix + 1)
            track.draw_track(sequence, ax[idx_track], chrom, sequence.sharex)
    if sequence.scale and sequence.sharex:
        ax[-1].ticklabel_format(axis="x", style="sci", scilimits=(0, 0))
        ax[-1].spines['bottom'].set_visible(True)
        ax[-1].get_xaxis().set_visible(True)
        ax[-1].tick_params(axis='x', labelsize=sequence.x_label_fontsize)


    logging.info("Exporting figure / layout rearrangments")

    plt.rcParams['savefig.facecolor'] = canvas.background_color
    fig.suptitle(canvas.title,fontsize=canvas.title_fontsize)
    #plt.legend(legend_line, legend_title,bbox_to_anchor=(1,1), fontsize=15)
    fig.savefig(args.output)


if __name__ == "__main__":
    sys.exit(main())



import sys

from distutils.core import setup, Extension


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


sys.path.insert(0, "karyoplot")
import version
current_version = version.__version__

setup(name='karyoplot',
      version = current_version,
      description='karyoplot: annotations along the genome',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Nicolas Lapalu',
      author_email='nicolas.lapalu@inrae.fr',
      url="https://forgemia.inra.fr/bioger/karyoplot",
      license_files = ('LICENSE',),
      install_requires=[
          'numpy',
          'pysam',
          'matplotlib',
          'pandas',
          ],
      entry_points = {
        'console_scripts': ['karyoplot=karyoplot.main:main'],
      },
      packages = ['karyoplot'],
      classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
      ]
      )
